from __future__ import annotations

import shutil
import webbrowser

import nox
from nox import Session

from exasol.toolbox.nox._shared import DOCS_OUTPUT_DIR
from noxconfig import (
    PROJECT_CONFIG,
    Config,
)


def _build_docs(session: nox.Session, config: Config) -> None:
    session.run(
        "poetry",
        "run",
        "sphinx-build",
        "-W",
        "-b",
        "html",
        f"{config.doc}",
        DOCS_OUTPUT_DIR,
    )


def _build_multiversion_docs(session: nox.Session, config: Config) -> None:
    session.run(
        "poetry",
        "run",
        "sphinx-multiversion",
        f"{config.doc}",
        DOCS_OUTPUT_DIR,
    )


@nox.session(name="docs:multiversion", python=False)
def build_multiversion(session: Session) -> None:
    """Builds the multiversion project documentation"""
    _build_multiversion_docs(session, PROJECT_CONFIG)


@nox.session(name="docs:build", python=False)
def build_docs(session: Session) -> None:
    """Builds the project documentation"""
    _build_docs(session, PROJECT_CONFIG)


@nox.session(name="docs:open", python=False)
def open_docs(session: Session) -> None:
    """Opens the built project documentation"""
    docs_folder = PROJECT_CONFIG.root / DOCS_OUTPUT_DIR
    if not docs_folder.exists():
        session.error(f"No documentation could be found. {docs_folder} is missing")
    index = docs_folder / "index.html"
    webbrowser.open_new_tab(index.as_uri())


@nox.session(name="docs:clean", python=False)
def clean_docs(_session: Session) -> None:
    """Removes the documentations build folder"""
    docs_folder = PROJECT_CONFIG.root / DOCS_OUTPUT_DIR
    if docs_folder.exists():
        shutil.rmtree(docs_folder)



from __future__ import annotations

import argparse
from pathlib import Path
from typing import (
    List,
    Tuple,
)

import nox
from nox import Session

from exasol.toolbox import cli
from exasol.toolbox.nox._shared import (
    Mode,
    _version,
)
from exasol.toolbox.nox.plugin import NoxTasks
from exasol.toolbox.release import (
    Version,
    extract_release_notes,
    new_changelog,
    new_changes,
    new_unreleased,
)
from noxconfig import PROJECT_CONFIG


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="nox -s release:prepare",
        usage="nox -s release:prepare -- [-h] version",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "version",
        type=cli.version,
        help=("A version string of the following format:" '"NUMBER.NUMBER.NUMBER"'),
    )
    parser.add_argument(
        "--no-add",
        default=False,
        action="store_true",
        help=("Neither add nor commit the changes"),
    )
    parser.add_argument(
        "--no-branch",
        default=False,
        action="store_true",
        help=("Do not create a branch to commit the changes on"),
    )
    parser.add_argument(
        "--no-pr",
        default=False,
        action="store_true",
        help=("Do not create a pull request for the changes"),
    )
    return parser


def _is_valid_version(old: Version, new: Version) -> bool:
    return new >= old


def _update_project_version(session: Session, version: Version) -> Version:
    session.run("poetry", "version", f"{version}")
    _version(session, Mode.Fix, PROJECT_CONFIG.version_file)
    return version


def _update_changelog(version: Version) -> tuple[Path, Path, Path]:
    unreleased = Path(PROJECT_CONFIG.root) / "doc" / "changes" / "unreleased.md"
    changelog = Path(PROJECT_CONFIG.root) / "doc" / "changes" / f"changes_{version}.md"
    changes = Path(PROJECT_CONFIG.root) / "doc" / "changes" / f"changelog.md"

    changelog_content = extract_release_notes(unreleased)
    changelog.write_text(new_changelog(version, changelog_content))

    unreleased.write_text(new_unreleased())

    changes_content = new_changes(changes, version)
    changes.write_text(changes_content)

    return changelog, changes, unreleased


def _add_files_to_index(session: Session, files: list[Path]) -> None:
    for file in files:
        session.run("git", "add", f"{file}")


@nox.session(name="release:prepare", python=False)
def prepare_release(session: Session, python=False) -> None:
    """
    Prepares the project for a new release.
    """
    parser = _create_parser()
    args = parser.parse_args(session.posargs)

    if not _is_valid_version(
        old=(old_version := Version.from_poetry()),
        new=(new_version := args.version),
    ):
        session.error(
            f"Invalid version: the release version ({new_version}) "
            f"must be greater than or equal to the current version ({old_version})"
        )

    if not args.no_branch and not args.no_add:
        session.run("git", "switch", "-c", f"release/prepare-{new_version}")

    pm = NoxTasks.plugin_manager(PROJECT_CONFIG)

    _ = _update_project_version(session, new_version)
    changelog, changes, unreleased = _update_changelog(new_version)

    pm.hook.prepare_release_update_version(
        session=session, config=PROJECT_CONFIG, version=new_version
    )

    if args.no_add:
        return

    files = [
        changelog,
        unreleased,
        changes,
        PROJECT_CONFIG.root / "pyproject.toml",
        PROJECT_CONFIG.version_file,
    ]
    results = pm.hook.prepare_release_add_files(session=session, config=PROJECT_CONFIG)
    files += [f for plugin_response in results for f in plugin_response]
    _add_files_to_index(
        session,
        files,
    )
    session.run("git", "commit", "-m", f"Prepare release {new_version}")

    if not args.no_pr:
        session.run(
            "gh",
            "pr",
            "create",
            "--title",
            f"Prepare release {new_version}",
            "--body",
            '""',
        )



from pathlib import Path
from typing import List


def update_workflow(template: Path, version: str) -> None:
    """Updates versions of XYZ in GitHub workflow ..."""
    with open(template, encoding="utf-8") as file:
        content = file.readlines()

    content = update_versions(
        lines=content, matcher="exasol/python-toolbox/.github/", version=version
    )

    with open(template, "w", encoding="utf-8") as file:
        file.writelines(content)


def is_update_required(line, matcher):
    return matcher in line and "@" in line


def update_version(line, version):
    keep = line[: line.index("@") + 1]
    updated = f"{version}\n"
    return f"{keep}{updated}"


def update_versions(lines, matcher, version) -> List[str]:
    result = []
    for line in lines:
        if is_update_required(line, matcher):
            line = update_version(line, version)
        result.append(line)
    return result



from __future__ import annotations

import subprocess
from dataclasses import dataclass
from datetime import datetime
from functools import total_ordering
from inspect import cleandoc
from pathlib import Path
from shutil import which

from exasol.toolbox.error import ToolboxError


def _index_or(container, index, default):
    try:
        return container[index]
    except IndexError:
        return default


@total_ordering
@dataclass(frozen=True)
class Version:
    major: int
    minor: int
    patch: int

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}"

    def __lt__(self, other: object):
        if not isinstance(other, Version):
            return NotImplemented
        return (
            self.major < other.major
            or (self.major <= other.major and self.minor < other.minor)
            or (
                self.major <= other.major
                and self.minor <= other.minor
                and self.patch < other.patch
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, Version):
            return NotImplemented
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    @staticmethod
    def from_string(version):
        parts = [int(number, base=0) for number in version.split(".")]
        if len(parts) > 3:
            raise ValueError(
                "Version has an invalid format, "
                f"expected: '<major>.<minor>.<patch>', actual: '{version}'"
            )
        version = [_index_or(parts, i, 0) for i in range(3)]
        return Version(*version)

    @staticmethod
    def from_poetry():
        poetry = which("poetry")
        if not poetry:
            raise ToolboxError("Couldn't find poetry executable")

        try:
            result = subprocess.run(
                [poetry, "version", "--no-ansi", "--short"], capture_output=True
            )
        except subprocess.CalledProcessError as ex:
            raise ToolboxError() from ex
        version = result.stdout.decode().strip()

        return Version.from_string(version)


def extract_release_notes(file: str | Path) -> str:
    """
    Extract release notes from a given file.

    Args:
        file: from which the release notes shall be extracted

    Returns:
        A string containing the cleaned release notes extracted from the file.
    """
    with open(file) as f:
        lines = f.readlines()[1:]
        content = "".join(lines)
        content = cleandoc(content)
    return content


def new_changes(file: str | Path, version: Version) -> str:
    """
    Create a new changelog list, adding the provided version to it.

    Args:
        file: containing the old changelog list.
        version: of the new entry to add to the list.

    Returns:
        Content for the new changelog list.
    """
    file = Path(file)
    content = []

    with open(file) as f:
        for line in f:
            content.append(line)
            if line.startswith("* [unreleased]"):
                content.append(f"* [{version}](changes_{version}.md)\n")
            if line.startswith("unreleased"):
                content.append(f"changes_{version}\n")

    return "".join(content)


def new_unreleased() -> str:
    """
    Generates the content for a new Unreleased changelog file.

    Returns:
        A string representing the content for an empty Unreleased changelog file.
    """
    return "# Unreleased\n"


def new_changelog(version: Version, content: str, date: datetime | None = None) -> str:
    """
    Create a changelog entry for a specific version.

    Args:
        version: An instance of the Version class representing the release version.
        content: The content of the changelog entry.
        date: Optional. The release date. If not provided, the current date will be used.

    Returns:
        The generated changelog.
    """
    date = datetime.today() if date is None else date
    template = cleandoc(
        """
        # {version} - {date}

        {content}
        """
    )
    return template.format(
        version=version, date=date.strftime("%Y-%m-%d"), content=content
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exasol',
 'exasol.toolbox',
 'exasol.toolbox.nox',
 'exasol.toolbox.pre_commit_hooks',
 'exasol.toolbox.release',
 'exasol.toolbox.sphinx',
 'exasol.toolbox.sphinx.multiversion',
 'exasol.toolbox.templates',
 'exasol.toolbox.tools']

package_data = \
{'': ['*'],
 'exasol.toolbox.sphinx.multiversion': ['templates/*'],
 'exasol.toolbox.templates': ['github/*',
                              'github/ISSUE_TEMPLATE/*',
                              'github/PULL_REQUEST_TEMPLATE/*',
                              'github/actions/python-environment/*',
                              'github/workflows/*']}

install_requires = \
['bandit[toml]>=1.7.9,<2.0.0',
 'black>=24.1.0',
 'coverage>=6.4.4,<8.0.0',
 'furo>=2022.9.15',
 'import-linter>=2.0,<3.0',
 'importlib-resources>=5.12.0',
 'isort>=5.12.0,<6.0.0',
 'jinja2>=3.1.4,<4.0.0',
 'mypy>=0.971',
 'myst-parser>=2.0.0,<4',
 'nox>=2022.8.7',
 'pluggy>=1.5.0,<2.0.0',
 'pre-commit>=3.1.1,<4',
 'prysk[pytest-plugin]>0.17.0,<1',
 'pylint>=2.15.4',
 'pytest>=7.2.2,<9',
 'pyupgrade>=2.38.2,<4.0.0',
 'shibuya>=2024.5.14',
 'sphinx-copybutton>=0.5.0,<0.6.0',
 'sphinx-design>=0.5.0,<1',
 'sphinx-inline-tabs>=2023.4.21,<2024.0.0',
 'sphinx>=5.3,<8',
 'typer[all]>=0.7.0']

entry_points = \
{'console_scripts': ['sphinx-multiversion = '
                     'exasol.toolbox.sphinx.multiversion:main',
                     'tbx = exasol.toolbox.tools.tbx:CLI',
                     'version-check = '
                     'exasol.toolbox.pre_commit_hooks.package_version:main']}

setup_kwargs = {
    'name': 'exasol-toolbox',
    'version': '0.17.0',
    'description': 'Your one-stop solution for managing all standard tasks and core workflows of your Python project.',
    'long_description': '<h1 align="center">Exasol Toolbox</h1>\n\n<p align="center">\nYour one-stop solution for managing all standard tasks and core workflows of your Python project.\n</p>\n\n<p align="center">\n\n<a href="https://github.com/exasol/python-toolbox/actions/workflows/ci.yml">\n    <img src="https://github.com/exasol/python-toolbox/actions/workflows/ci.yml/badge.svg?branch=main" alt="Checks Main">\n</a>\n<a href="https://opensource.org/licenses/MIT">\n    <img src="https://img.shields.io/pypi/l/exasol-toolbox" alt="License">\n</a>\n<a href="https://pypi.org/project/exasol-toolbox/">\n    <img src="https://img.shields.io/pypi/dm/exasol-toolbox" alt="Downloads">\n</a>\n<a href="https://pypi.org/project/exasol-toolbox/">\n    <img src="https://img.shields.io/pypi/pyversions/exasol-toolbox" alt="Supported Python Versions">\n</a>\n<a href="https://pypi.org/project/exasol-toolbox/">\n    <img src="https://img.shields.io/pypi/v/exasol-toolbox" alt="PyPi Package">\n</a>\n</p>\n\n## 🚀 Features\n\n- Centrally managed standard tasks\n  - code formatting & upgrading\n  - linting\n  - type-checking\n  - unit-tests\n  - integration-tests\n  - coverage\n  - documentation\n\n- Centrally manged core workflows\n  - workspace/project verification\n  - build and publish releases\n  - build and publish documentation\n\n- Configurable & Extensible\n  - Project configuration\n  - Event hooks\n\n## 🔌️ Prerequisites\n\n- [Python](https://www.python.org/) >= 3.8\n\n## 💾 Installation\n\n```shell\npip install exasol-toolbox\n```\n\n## 📚 Documentation\n\nFor futher details, checkout the latest [documentation](https://exasol.github.io/python-toolbox/).\n\n',
    'author': 'Nicola Coretti',
    'author_email': 'nicola.coretti@exasol.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
